'''

 Filter to create a NEW image applying a pseudo-hdr effect to the INPUT DRAWABLE
 Follows the tut of jneurock@gimpology.com, here the link :
 http://gimpology.com/submission/view/fake_hdr_look_in_gimp
 with modifications by dinasset
 ..........................................................................................
 
 Prerequisites:
 - script-fu-invert-YUV by Stratadrake available at http://registry.gimp.org/node/8352
 - the ancillary function "WhiteBalanceStretch" available here under Gimp Scripts & Plug-ins 
   (used only if WhiteBalance has been requested) 
 ..........................................................................................   
 
 Release 1.0   initial
 
 Flow implemented to get the final result:      ...detail the major steps
 
1. save current context
 2. prepare the new image, copying the source drawable into the new BG (dimensions as user option)
 3. apply a White Balance Stretch (if requested by the user: uses a separate plug-in)
 4. duplicate the BG, name it DesatInvertedLayer
    desaturate, invert luma using script-fu-invert-YUV, apply plug-in SoftGlow,
    set mode = SOFTLIGHT at 55% 
 5. duplicate the DesatInvertedLayer,
    set mode = SOFTLIGHT at 82%
 6. duplicate the BG, name it LevelALayer
    set mode = HARDLIGHT at 35%
    stretch Level values from 96-255 to 0-255
 7. create a new layer from visible (VisibleA)
 8. duplicate VisibleA, name it DBRMlayer
 9. if DodgeBurn has been selected:
        use the service routine DoDodgeBurn to create a Dodge Channel
        use it again to create a Burn Channel
        merge all into the DBRMlayer
        (logic for the dodge and burn function is a translation from .scm to .py
        of the existing filter script-fu-dodge-burn by Harry Phillips)
 10.if Retinex has been selected:
        duplicate DBRMlayer
        apply Retinex
        desaturate
        set mode = SOFTLIGHT at 100%
        merge down into the DBRMlayer
 11.apply a compress level values from 0-225 to 8-247 (suppress extremes)
 12.set mode = SCREEN at 35%
 13.new layer from visible (FinalTop)
    apply the requested Saturation change
    if Denoise has been selected apply G'MIC Iain fast denoise
    apply a curve to let values stay between 8 and 247
 14.flatten the new image if requested
 
 
 ----------------------------------------------------------------
 COPYRIGHT NOTICE
 ----------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 -------------------------------------------------------------------
 other info
 -------------------------------------------------------------------
  
'''
#!/usr/bin/env python
from gimpfu import *
import decimal


############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
############################################################################
#

# ---------------------------------------------------------------------------

def DodgeBurn (myImage, myLayer, modeOp, thinNum, thickNum) :

    BaseLayer = pdb.gimp_layer_new_from_drawable(myLayer, myImage)
    myImage.add_layer (BaseLayer, 0)
    BaseLayer.name = "First Work Layer"
    pdb.gimp_desaturate (BaseLayer)
    
    ThinLayer = BaseLayer.copy()
    myImage.add_layer (ThinLayer, 0)
    ThinLayer.name = "Thin Work Layer"
    
    pdb.plug_in_gauss (myImage, ThinLayer, thinNum, thinNum, 0)
    
    if modeOp == 1 :
        ThinLayer.mode = LIGHTEN_ONLY_MODE
    else :
        ThinLayer.mode = DARKEN_ONLY_MODE
    
    pdb.plug_in_gauss (myImage, BaseLayer, thickNum, thickNum, 0)
    ThickLayer = BaseLayer.copy()
    myImage.add_layer (ThickLayer, 1)
    ThickLayer.name = "Thick Work Layer"
 
    DBLayer = pdb.gimp_image_merge_down (myImage, ThinLayer, 0) 
    DBLayer.mode = DIFFERENCE_MODE
    DBLayer = pdb.gimp_image_merge_down (myImage, DBLayer, 0) 
 
    if modeOp == 1 :
        DBLayer.mode = DODGE_MODE
        DBLayer.name = "DODGE CHANNEL"
    else :
        DBLayer.mode = BURN_MODE
        DBLayer.name = "BURN CHANNEL"
        pdb.gimp_invert (DBLayer)
 
    return (DBLayer)

# ------------------------------------------------------------------------------


def PseudoHdrEffectSE (inImage, inDrawable, widthheight,
                     doStretch, saturat,
                     doDodgeBurn, doRetinexMix, doDenoise, doFlatten) :

    pdb.gimp_context_push
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    width = inImage.width
    height = inImage.height
    type = inImage.base_type
    
    name = inDrawable.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    
    # calculate scale factor for the new output image
    if widthheight == 0 :   # means the user wants to have same size as input
        scale = 1
    else :     
        optiWidthHeight = widthheight       # size to work on to get best results
        realWidthHeight = width+height      # real size of the input image
        scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling

    # Set up an undo group, so the operation will be undone in one step.
    pdb.gimp_image_undo_group_start(inImage)
    # fix all required context settings 
    pdb.gimp_context_set_antialias(True)
    pdb.gimp_context_set_feather(False)
    pdb.gimp_context_set_feather_radius(0, 0)
    pdb.gimp_context_set_sample_merged(False) 
    pdb.gimp_context_set_sample_criterion(SELECT_CRITERION_COMPOSITE) 
    pdb.gimp_context_set_sample_threshold(0.00)
    pdb.gimp_context_set_sample_transparent(False)       # beware, if TRUE may change the area selected!
    pdb.gimp_context_set_interpolation(INTERPOLATION_LANCZOS)
    
    newwidth=int(width*scale)
    newheight=int(height*scale)
    
    #info_msg ("newwidth="+str(newwidth))
    newImage = pdb.gimp_image_new (newwidth, newheight, type)                    
    newBG = pdb.gimp_layer_new_from_drawable(inDrawable, newImage)     # background layer
    newBG.mode = NORMAL_MODE
    newBG.name = "Background Image Layer"
    newBG.opacity = 100.0 
    newImage.add_layer (newBG, 0)
    pdb.gimp_layer_scale(newBG, newwidth, newheight, False)
    alpha=pdb.gimp_drawable_has_alpha(newBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (newBG)    #  RGBA
    if (doStretch == True) :
        # stretching the colours 
        newBG = pdb.python_fu_WhiteBalanceStretch (newImage, newBG)
        newBG.name = "Background Image Layer, stretched"
    #pdb.gimp_display_new(newImage)    
    #  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< 
    DesatInvertedLayer = newBG.copy()
    DesatInvertedLayer.mode = SOFTLIGHT_MODE
    DesatInvertedLayer.name = "Desaturated Inverted Layer"
    DesatInvertedLayer.opacity = 55.0
    newImage.add_layer (DesatInvertedLayer, 0)
    DesatInvertedLayer = newImage.active_layer
    pdb.gimp_desaturate_full (DesatInvertedLayer, 2)
    #pdb.gimp_invert (DesatInvertedLayer) #if no script available for luma invert
    pdb.script_fu_invert_YUV (inImage, DesatInvertedLayer)
    radius = 10.00
    bright = 0.75
    sharpn = 0.85
    pdb.plug_in_softglow (inImage, DesatInvertedLayer, radius, bright, sharpn)                        
    
    DesatInvertedLayer2 = DesatInvertedLayer.copy()
    DesatInvertedLayer2.mode = SOFTLIGHT_MODE
    DesatInvertedLayer2.name = "2.nd Desaturated Inverted Layer"
    DesatInvertedLayer2.opacity = 82.0 
    newImage.add_layer (DesatInvertedLayer2, 0)
    DesatInvertedLayer2 = newImage.active_layer
    
    LevelALayer = newBG.copy()
    LevelALayer.mode = HARDLIGHT_MODE
    LevelALayer.name = "Level Layer"
    LevelALayer.opacity = 35.0
    newImage.add_layer (LevelALayer, 0)
    LevelALayer = newImage.active_layer
    minvalue = 96
    pdb.gimp_levels (LevelALayer, HISTOGRAM_VALUE, minvalue, 255, 1.0, 0, 255)
    
    VisibleA = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible A")
    VisibleA.name = "Visible after step A"
    newImage.add_layer (VisibleA, 0)
    
    DBRMlayer = VisibleA.copy()
    DBRMlayer.name = "Layer for extra effects"
    newImage.add_layer (DBRMlayer, 0)
    DBRMlayer.mode = NORMAL_MODE
    DBRMlayer.opacity = 100.0
    DBRMlayer = newImage.active_layer

    if doDodgeBurn == True :
        thin = (newwidth+newheight)/250
        thick = (newwidth+newheight)/100
        dodgeL = 1
        burnL = 0
        # Do the dodge layer first
        DBD = DodgeBurn (newImage, DBRMlayer, dodgeL, thin, thick)
        pdb.gimp_image_set_active_layer (newImage, DBRMlayer)
        # ...then do the burne layer
        DBB = DodgeBurn (newImage, DBRMlayer, burnL, thin, thick)
        pdb.gimp_image_set_active_layer (newImage, DBRMlayer)
        # merge down
        MergeDBD = pdb.gimp_image_merge_down (newImage, DBD, EXPAND_AS_NECESSARY)
        MergeDBB = pdb.gimp_image_merge_down (newImage, DBB, EXPAND_AS_NECESSARY)
        DBRMlayer = newImage.active_layer
        DBRMlayer.name = "Layer for extra effects: DodgeBurne applied"
        
    if doRetinexMix == True :
        DBRMlayer2 = DBRMlayer.copy()
        DBRMlayer2.name = "Layer for extra effect Retinex"
        newImage.add_layer (DBRMlayer2, 0)
        DBRMlayer2.opacity = 100.0
        DBRMlayer2 = newImage.active_layer
        SCALE=240
        NSCALES=3
        SCALES_MODE=0   # uniform
        CVAR=1.2
        pdb.plug_in_retinex(inImage, DBRMlayer2, SCALE, NSCALES, SCALES_MODE, CVAR)
        pdb.gimp_desaturate_full (DBRMlayer2, DESATURATE_LUMINOSITY)
        DBRMlayer2.mode = SOFTLIGHT_MODE
        MergeDBRM2 = pdb.gimp_image_merge_down (newImage, DBRMlayer2, EXPAND_AS_NECESSARY)

        DBRMlayer = newImage.active_layer
        if doDodgeBurn == True :
            DBRMlayer.name = "Layer for extra effects: DodgeBurne & Retinex applied"
        else :    
            DBRMlayer.name = "Layer for extra effects: RetinexMix applied"
    
    #pdb.gimp_display_new(newImage)
    #return
    
    DBRMlayer = newImage.active_layer
    pdb.gimp_hue_saturation (DBRMlayer, 0, 0, 0, saturat)
    fromlev = 8
    tolev = 247
    #pdb.gimp_levels (DBRMlayer, HISTOGRAM_VALUE, fromlev, tolev, 1.0, 0, 255)
    pdb.gimp_levels (DBRMlayer, HISTOGRAM_VALUE, 0, 255, 1.0, fromlev, tolev)
    pdb.gimp_image_set_active_layer (newImage, DBRMlayer)
    DBRMlayer.mode = SCREEN_MODE
    DBRMlayer.opacity = 35.00
    
    FinalTopLayer = pdb.gimp_layer_new_from_visible (newImage, newImage, "Top visible")
    newImage.add_layer (FinalTopLayer, 0)
    FinalTopLayer.mode = NORMAL_MODE
    FinalTopLayer.opacity = 100.0
    FinalTopLayer = newImage.active_layer
    pdb.gimp_hue_saturation (FinalTopLayer, 0, 0, 0, saturat)
    pdb.gimp_brightness_contrast (FinalTopLayer, 0, 5)
    pdb.plug_in_unsharp_mask (newImage, FinalTopLayer, 3.0, 0.3, 0)
    FinalTopLayer.name = "Top visible, saturation at "+str(saturat)
    if doDenoise == True :
        ####################################
        # replaceed in the Special Edition #
        ####################################
        #[G'MIC] Iain's fast denoise:  
        ###pdb.plug_in_gmic(newImage, FinalTopLayer, 1, 
        ###           "-v - " 
        ###           +"-iain_fast_denoise_p 1.08,1.08,1,0.10,0,0,1 ")
        pdb.plug_in_sel_gauss (newImage, FinalTopLayer, 9.0, 50)
        
    pdb.gimp_curves_spline (FinalTopLayer, HISTOGRAM_VALUE, 6,
                            (0, 8,  127, 127, 255, 247) 
                           )
    
    if doFlatten == True :
        pdb.gimp_image_flatten(newImage)
        flattenedLayer = pdb.gimp_image_get_active_layer (newImage)
        if doStretch == True :
            if doDodgeBurn == True :
                if doRetinexMix == True :
                    flattenedLayer.name = "PseudoHDR on "+name+" stretch, dodge-burn, retinex, satur at "+str(saturat)
                else :
                    flattenedLayer.name = "PseudoHDR on "+name+" stretch, dodge-burn, satur at "+str(saturat)
            else :
                if doRetinexMix == True :
                    flattenedLayer.name = "PseudoHDR on "+name+" stretch, retinex, satur at "+str(saturat)
                else :
                    flattenedLayer.name = "PseudoHDR on "+name+" stretch, satur at "+str(saturat)
        else :
            if doDodgeBurn == True :
                if doRetinexMix == True :
                    flattenedLayer.name = "PseudoHDR on "+name+", dodge-burn, retinex, satur at "+str(saturat)
                else :
                    flattenedLayer.name = "PseudoHDR on "+name+" dodge-burn, satur at "+str(saturat)
            else :
                if doRetinexMix == True :
                    flattenedLayer.name = "PseudoHDR on "+name+" retinex, satur at "+str(saturat)
                else :
                    flattenedLayer.name = "PseudoHDR on "+name+" satur at "+str(saturat)
    
    pdb.gimp_display_new(newImage)
    
    pdb.gimp_displays_flush
   
    # Close the undo group.
    pdb.gimp_image_undo_group_end(inImage)
    
    pdb.gimp_context_pop
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)
    
    if doFlatten == True :
        return(flattenedLayer)
    else :
        return()


# This is the plugin registration function
register(
    "PseudoHdrEffectSE",    
    "to create a new image applying a pseufo-hdr effect to the source drawable (Special Edition)",   
    "This script creates a new image applying a pseufo-hdr effect to the source drawable (Special Edition)",
    "Diego", 
    "Diego Nassetti ", 
    "August 2016",
    "PseudoHdrEffectSE", 
    "RGB*", 
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "idrawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image (0=same as input)", 0, (0, 9600, 300)),
      (PF_TOGGLE, "stretch", "White Balance?", True),
      (PF_ADJUSTMENT, "satur", "Saturation Adjustment", 15, (-100, 100, 1)),
      (PF_TOGGLE, "dodgeburn", "Add DodgeBurn effect?", True),
      (PF_TOGGLE, "retinex", "Add RetinexMix effect?", True),
      (PF_TOGGLE, "denoise", "Denoise Final outcome? (good for people)", False),
      (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ], 
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),
    ],
    PseudoHdrEffectSE,
    menu="<Image>/Diego/Enhance",
    )

main()
